#!/bin/bash
########################################################################
# build_root.sh
#
# Build root and package for UPS.
#
# Adapted from build.sh.template provided by ssibuildshims v2_03_00.
########################################################################

########################################################################
# Non-customize-able boilerplate: skip to "Customize-able functions."
########################################################################

####################################
# Useful variables.
prog=${BASH_SOURCE[0]##*/}

##################
# Local (dumb) ssi_die() function for use until we source ssi_functions.
ssi_die() {
  printf "FATAL ERROR: ${*}\n" 1>&2; exit 1
}
##################

##################
# Basic usage function for use by ensure_buildshims until we have enough
# information to provide something more specific. Do NOT customize!
build_script_usage() {
    cat <<EOF
USAGE: ${prog} <ssi-msg-config-options> [--] <product-dir> <qualifier-args> <maketar>
       ${prog} -h|--help [<product-dir>]

  Build ${package} and package for UPS.

OPTIONS

  -h|--help

    Help. Supply <product-dir> for specific, detailed information on
    invoking ${prog}.

  <ssi-msg-config-options>

    Supply <product-dir> for more information.

ARGUMENTS

  <product-dir>

    The top-level UPS directory in which the source code distribution
    was unpacked. When specified with -h|--help, this enables us to
    provide more specific help about invoking ${prog}.

  <qualifier-args>
  <maketar>

    Supply <product-dir> for more information.

EOF
}
##################

##################
# Make sure we have UPS and ssibuildshims set up.
ensure_ssibuildshims() {
  if ! (( $# )) || [ "${@: -1}" = "-h" ] || [ "${@: -1}" = "--help" ]; then
    build_script_usage
    exit 1
  fi
  # Find <product-dir> from our arguments:
  local arg
  for arg in "${@}"; do
    [[ "$arg" ==  -* ]] && continue;
    product_dir="${arg}"; break
  done
  if [ -z "${product_dir}" ]; then
    # Bad usage.
    ssi_die "Bad <product-dir> ${product_dir}\n$(build_script_usage)"
    exit 2
  elif ! { [ -d "${product_dir}/.upsfiles" ] && [ -w "${product_dir}" ]; }; then
    # Something wrong with <product-dir>.
    ssi_die "<product-dir> not a valid, writable unified UPS directory: " \
      "${product_dir}\n$(build_script_usage)"
  fi
  # Promising...
  local abs_product_dir="$(cd "${product_dir}" 2>/dev/null 2>&1 && pwd -P)"
  (( $? )) && ssi_die "unable to cd to <product-dir>: ${product_dir}\n" \
    "$(build_script_usage)"
  # Find our own bootstraps and yank 'em.
  local ssibuildshims_topdir="${product_dir}/ssibuildshims/${ssibuildshims_version}"
  [ -f "${ssibuildshims_topdir}/bin/ssi_functions" ] && \
    source "${ssibuildshims_topdir}/bin/ssi_functions" >/dev/null 2>&1 && \
    ensure_ups "${product_dir}" && \
    setup ssibuildshims ${ssibuildshims_version} -z "${product_dir}" || \
    ssi_die "unable to find ssibuildshims ${ssibuildshims_version} in
${product_dir} and set it up with UPS."
  # Absolute is safer.
  absolute_path -- "${product_dir}" product_dir
}
##################

########################################################################
# Customize-able functions.
########################################################################

##################
# Ascertain whether we built and installed correctly. Add more /
# different calls to verify_in_filesystem as appropriate. Arguments
# (e.g. -q) are passed through to verify_in_filesystem.
is_built() { # *C*
  ${trace} verify_in_filesystem -d "${@}" "${pkgdir}/lib" && \
    ${trace} verify_in_filesystem -x "${@}" "${pkgdir}/bin/root.exe" && \
    ${trace} verify_in_filesystem -x "${@}" "${pkgdir}/bin/root"
}
##################

##################
# If a distribution archive is wanted and the installation is successful
# according to is_built(), then make it. Any arguments (e.g. -q) are
# passed through both to is_built() and make_basic_distribution().
#
# If make_basic_distribution() finds that the product is declared
# correctly and attempts to make the distribution, then it will cause us
# to exit with the status code of that operation. If not, it returns an
# error code.
#
# See make_basic_distribution -h for useful options to add to
# mbd_opts(). *C*
maybe_make_distribution() {
  if [ "${maketar}" = "tar" ]; then
    # Customize mbd_opts. *C*
    local mbd_opts=()
    ${trace} is_built "$@" && \
      ${trace} make_basic_distribution "${@}" "${mbd_opts[@]}"
  fi
}
##################

########################################################################
# Package-specific functions.
########################################################################

# Steering options.
add_steering_options() {
  local rcxxmodules=ON
  (( ROOT_NO_CXXMODULES )) && rcxxmodules=OFF

  cmake_options+=( \
    -Dall:BOOL=OFF \
    -Dccache:BOOL=OFF \
    -Dclingtest:BOOL=OFF \
    -Dcoverage:BOOL=OFF \
    -Dcxxmodules:BOOL=OFF \
    -Dexceptions:BOOL=ON \
    -Dfail-on-missing:BOOL=ON \
    -Dfortran:BOOL=ON \
    -Dgminimal:BOOL=ON \
    -Dgnuinstall:BOOL=OFF \
    -Dmemory_termination:BOOL=OFF \
    -Dminimal:BOOL=OFF \
    -Drootbench:BOOL=OFF \
    -Droottest:BOOL=OFF \
    -Drpath:BOOL=OFF \
    -Druntime_cxxmodules:BOOL=${rcxxmodules} \
    -Dshared:BOOL=ON \
    -Dsoversion:BOOL=OFF \
    -Dtesting:BOOL=OFF \
    )

  [ "${cmake_build_type}" = "Debug" ] && \
    (( ROOT_DEBUG_LLVM )) && \
    cmake_options+=( \
    "-DLLVM_BUILD_TYPE=${cmake_build_type}" \
    "-DLLVM_ENABLE_ASSERTIONS:BOOL=ON" \
    )
}

# Builtin options.
add_builtin_options() {
  cmake_options+=( \
    -Dbuiltin_afterimage:BOOL=OFF \
    -Dbuiltin_cfitsio:BOOL=OFF \
    -Dbuiltin_clang:BOOL=ON \
    -Dbuiltin_davix:BOOL=OFF \
    -Dbuiltin_fftw3:BOOL=OFF \
    -Dbuiltin_freetype:BOOL=OFF \
    -Dbuiltin_ftgl:BOOL=OFF \
    -Dbuiltin_gl2ps:BOOL=OFF \
    -Dbuiltin_glew:BOOL=OFF \
    -Dbuiltin_gsl:BOOL=OFF \
    -Dbuiltin_llvm:BOOL=ON \
    -Dbuiltin_lz4:BOOL=OFF \
    -Dbuiltin_lzma:BOOL=OFF \
    -Dbuiltin_openssl:BOOL=OFF \
    -Dbuiltin_pcre:BOOL=OFF \
    -Dbuiltin_tbb:BOOL=OFF \
    -Dbuiltin_unuran:BOOL=OFF \
    -Dbuiltin_vc:BOOL=OFF \
    -Dbuiltin_vdt:BOOL=OFF \
    -Dbuiltin_veccore:BOOL=OFF \
    -Dbuiltin_xrootd:BOOL=OFF \
    -Dbuiltin_xxhash:BOOL=OFF \
    -Dbuiltin_zlib:BOOL=OFF \
    -Dbuiltin_zstd:BOOL=OFF \
    )
}

# Feature options and configuration.
add_feature_options() {
  cmake_options+=( \
    -DXROOTD_ROOT_DIR="${XROOTDROOT}" \
    -Dalien:BOOL=OFF \
    -Darrow:BOOL=OFF \
    -Dasimage:BOOL=ON \
    -Dcefweb:BOOL=OFF \
    -Dclad:BOOL=ON \
    -Dcuda:BOOL=OFF \
    -Dcudnn:BOOL=OFF \
    -Ddataframe:BOOL=ON \
    -Ddavix:BOOL=OFF \
    -Ddcache:BOOL=OFF \
    -Dfcgi:BOOL=OFF \
    -Dfftw3:BOOL=ON \
    -Dfitsio:BOOL=OFF \
    -Dgdml:BOOL=ON \
    -Dgfal:BOOL=OFF \
    -Dgsl_shared:BOOL=ON \
    -Dgviz:BOOL=OFF \
    -Dhttp:BOOL=ON \
    -Dimt:BOOL=ON \
    -Djemalloc:BOOL=OFF \
    -Dmathmore:BOOL=ON \
    -Dminuit2:BOOL=ON \
    -Dmlp:BOOL=ON \
    -Dmonalisa:BOOL=OFF \
    -Dmpi:BOOL=OFF \
    -Dmysql:BOOL=ON \
    -Dodbc:BOOL=OFF \
    -Dopengl:BOOL=ON \
    -Doracle:BOOL=OFF \
    -Dpgsql:BOOL=ON \
    -Dpyroot:BOOL=ON \
    -Dpyroot_experimental:BOOL=OFF \
    -Dpythia6:BOOL=ON \
    -Dpythia6_nolink:BOOL=OFF \
    -Dpythia8:BOOL=ON \
    -Dqt5web:BOOL=OFF \
    -Dr:BOOL=OFF \
    -Droofit:BOOL=ON \
    -Droot7:BOOL=OFF \
    -Dshadowpw:BOOL=OFF \
    -Dspectrum:BOOL=OFF \
    -Dsqlite:BOOL=OFF \
    -Dssl:BOOL=ON \
    -Dtcmalloc:BOOL=OFF \
    -Dtmva-cpu:BOOL=ON \
    -Dtmva-gpu:BOOL=OFF \
    -Dtmva-pymva:BOOL=ON \
    -Dtmva-rmva:BOOL=OFF \
    -Dtmva:BOOL=ON \
    -Dunuran:BOOL=OFF \
    -Dvc:BOOL=OFF \
    -Dvdt:BOOL=OFF \
    -Dveccore:BOOL=OFF \
    -Dvecgeom:BOOL=OFF \
    -Dwebgui:BOOL=ON \
    -Dxml:BOOL=ON \
    -Dxrootd:BOOL=ON \
    )
}

# Deal with compiler-specific options.
add_compiler_options() {
  local cc cstd cxx cxxstd fc fstd
  case ${cqual} in
    e19|e20)
      cc=gcc
      cstd=11
      cxx=g++
      cxxstd=17
      fc=gfortran
      fstd=legacy
      ;;
    e21)
      cc=gcc
      cstd=11
      cxx=g++
      cxxstd=20
      fc=gfortran
      fstd=legacy
      # Pythia6 and MySQL UPS builds do not yet support GCC
      # 10.1/C++2a/20.
      cmake_options+=(-Dpythia6:BOOL=OFF \
        -Dpythia6_nolink:BOOL=OFF \
        -Dmysql:BOOL=OFF \
        )
      ;;
    c7)
      cc=clang
      cstd=11
      cxx=clang++
      cxxstd=17
      fc=gfortran
      fstd=legacy
      if [ -z "${ROOT_NO_CXXMODULES+x}" ]; then
        # If we haven\'t explicitly set this variable (e.g. to empty) in
        # the environment, set it to 1 here.
        (( ROOT_NO_CXXMODULES = 1 ))
      fi
      extra_compiler_options=( \
        -DCLANG_DEFAULT_CXX_STDLIB=libc++ \
        -DLLVM_ENABLE_LIBCXX:BOOL=ON \
        -Dlibcxx:BOOL=ON \
        )
      ;;
    *) ssi_die "unrecognized compiler qualifier ${cqual}\n$(usage)"
  esac
  # No CMAKE_Fortran_STANDARD yet.
  fflg=("-std=legacy")
  cmake_options+=( \
    -DCMAKE_CXX_COMPILER=${cxx} \
    -DCMAKE_CXX_STANDARD=${cxxstd} \
    -DCMAKE_C_COMPILER=${cc} \
    -DCMAKE_C_STANDARD=${cstd} \
    -DCMAKE_Fortran_COMPILER=${fc} \
    -DCMAKE_Fortran_FLAGS="${fflg[@]}" \
    "${extra_compiler_options[@]}"
  )
  { ssi_indent_msg -f "Building with: " - | ssi_info -i -; } <<EOF
Building with: ${cc} -std=c${cstd} ($(eval which ${cc})
${cxx} -std=c++${cxxstd} ($(eval which ${cxx})
${fc} -std=${fstd} ($(eval which ${fc})

EOF
}

# Deal with platform-specific options.
add_platform_options() {
  if [ ${OS1} = "Darwin" ]; then
    cmake_options+=( \
      -Dbuiltin_afterimage:BOOL=ON \
      -Dbuiltin_freetype:BOOL=ON \
      -Dbuiltin_ftgl:BOOL=ON \
      -Dbuiltin_glew:BOOL=ON \
      -Dbuiltin_lzma:BOOL=ON \
      -Dbuiltin_xxhash:BOOL=ON \
      -Dbuiltin_zstd:BOOL=ON \
      -Dbuiltin_pcre:BOOL=ON \
      -Dcocoa:BOOL=ON \
      -Dmacos_native:BOOL=OFF \
      -Dx11:BOOL=OFF \
      )
    # Meant to be found only on El Capitan and newer.
    if [[ "${OS1}" == Darwin ]]; then
      os_base="${OS%-*}"
      if [[ -d /usr/local/opt/openssl ]] && \
        (( ${os_base#d} >= 15 )); then
        cmake_options+=(-DOPENSSL_ROOT_DIR=/usr/local/opt/openssl)
      fi
    fi
    if [[ "${cc}" = "gcc" ]]; then
      # GCC makes zero-size objects when __builtin_unreachable is used,
      # which panics the OS X linker.
      export CXXFLAGS="-D__builtin_unreachable=__builtin_trap"
      export CFLAGS="-D__builtin_unreachable=__builtin_trap"
    fi
  else
    if [[ $flvr == "Linux64bit+2.6-2.5" ]]; then # SLF 5.
      # LZMA only available through DAG -- not an option.
      cmake_options+=(-Dbuiltin_lzma:BOOL=ON)
    fi
  fi
}

########################################################################
# Main.
########################################################################

####################################
# Useful variables. *C*
trace=maybe_xtrace # Could set to xtrace instead for unconditional.
(( ssi_trace = 1 )) # Trace certain commands by default.
####################################

####################################
# Package name and version information. *C*
package=root
origpkgver=v6_20_04
pkgver=${origpkgver}b # Append letters for local patches.
ssibuildshims_version=v2_03_02
pkgdotver=${origpkgver//_/.}; pkgdotver=${pkgdotver#v}
####################################

##################
# Other useful variables. *C*
pkgtarname=${package}-${pkgdotver}
pkgtarfile=${pkgtarname}.tar.gz
##################

####################################
# Handle options and arguments.

# Set up ssibuildshims and UPS / product_dir as early as possible.
ensure_ssibuildshims "${@}"

##################
# Now parse options and arguments, setting basequal, extraqual, maketar,
# etc.
#
# Use verify_quals() to parse command line options, verify that the user
# has provided qualifiers supported by this build script, and set
# expected variables such as basequal, extraqual, cqual, etc.
#
# Customize the extra usage text as appropriate. Note that DEBUG_PATCHES
# and VERBOSE_PATCHES are honored by prepare_source() and
# patch_source(). DEBUG_(CONFIG|BUILD|TEST|INSTALL) should be accounted
# for (or not) by the relevant command in this script.
#
# See verify_quals -h for options and examples.
verify_quals \
  --cqual={e19,e20,e21,c7} \
  --btype={debug,prof} \
  --pqual=p383a \
  --parse --usage-file=- -- "$@" <<EOF || exit # *C*
ENVIRONMENT

  DEBUG_BUILD

    Verbose output from the corresponding step.

  DEBUG_PATCHES
  VERBOSE_PATCHES

    Command tracing and increased verbosity for source unpacking and patching
    operations. DEBUG_PATCHES will cause exit after patches have been applied.

  ROOT_DEBUG_LLVM=1

    When building for debug, additionally add debugging symbols and
    assertions to ROOT's internal LLVM. This will increase build time
    significantly, and run time somewhat.

  ROOT_DEBUG_ROOTCLING=1

    Increase verbosity for rootcling invocations during the ROOT build.

  ROOT_NO_CXXMODULES=1

    Deactivate C++ module generation during the ROOT build (activated by
    default since 6.20/00). Note that some platforms may have C++ module
    generation deactivated by default, in which case an explicit
    ROOT_NO_CXXMODULES=0 would be necessary to enable C++ module
    generation.

  ROOT_USE_NINJA=1

    Support for Fortran modules was added to Ninja in 1.10, and to CMake
    in 3.17. However, this may still be problematic, so we default to
    using GNU make. To use Ninja for the ROOT build, ensure Ninja >=
    1.10 and CMake >=3.17 are both available, and set ROOT_USE_NINJA=1.

EOF
##################

##################
# Define globals: pkgdir, sourcedir, blddir, fullqual, etc. See
# define_basics -h for steering options.
${trace} define_basics TRACK_PROGRESS || exit # *C*
##################

####################################
# Declare ourselves privately, and then set ourselves up to allow setup
# of dependencies. *C*
#
# Comment or delete if not required:
${trace} fake_declare_product -x -- || exit
####################################

####################################
# Build-only dependencies. *C*

# Require CMake.
${trace} setup_cmake --latest  || exit

# We can use GNU Make, but we'll take Ninja *if* we want it *and* it's
# available *and* it and CMake are modern enough to support Fortran.
(( ROOT_USE_NINJA )) && ! ver_greater 3.17.0 ${CMAKE_VERSION} && \
  ${trace} setup_ninja --optional --latest={--ge,1.10.0} && \
  (( want_ninja = 1 ))
####################################

# Print active products.
echo
ups active
echo

# Possible short-circuit if we're already built and declared.
${trace} maybe_make_distribution --silent

ssi_info "building ${package} for ${OS}-${plat}-${qualdir} (flavor ${pkgflvr})"

# Ensure directories: pkgdir, sourcedir, blddir.
${trace} init_pkg_dirs || exit

####################################
# Prepare source for building. *C*

${trace} cd "${sourcedir}" \
  || ssi_die "unable to cd to source directory ${sourcedir}"

# Unpack and patch the source (version-specific patch file is optional).
required_patches=( "fermi_root_6.patch" \
  "fermi_PrintArrayTemplateArgs.patch" \
  "fermi_compression-rootrc.patch" \
  "fermi_xrootd-rootrc.patch" \
)

# If we need to debug rootcling during the build.
(( ROOT_DEBUG_ROOTCLING )) && required_patches+=("rootcling-verbose.patch")

##################
# If we fail due to an incomplete previous source preparation, or
# updated patches or source are detected, make one attempt to recover
# automatically.
(( source_attempts = 0 ))
while true; do
  ${trace} prepare_source "${ssi_args[@]}" -P "${patchdir}" \
    -o "fermi_root_${pkgdotver//./_}.patch" -- \
    "${tardir}/${pkgtarfile}" \
    "${required_patches[@]}"
  case $? in
    0) break;;
    [34]) :;; # Incomplete or outdated remains.
    *) ssi_die "executing prepare_source"
  esac
  (( source_attempts++ == 0 )) || ssi_die "retry of prepare_source failed"
  ${trace} init_pkg_dirs -c || ssi_die "unable to clean up for retry."
done
##################

${trace} prepare_source "${ssi_args[@]}" -P "${patchdir}" \
  "${tardir}/${pkgtarfile}" \
  || exit

# Bail early if DEBUG_PATCHES is set.
[ -n "${DEBUG_PATCHES}" ] && exit 1
####################################

####################################
# Prepare to build. *C*
#
# Set configure / compile flags, environment variables, etc.

# Later options (e.g. compiler-specific ones) can override earlier ones.
declare -a cmake_options

# Determine and save the CMake build type, since we need it in a couple
# of places.
cmake_build_type=$(cmake_build_type "${extraqual}")

# Main CMake options first.
(( want_ninja )) && cmake_options+=(-GNinja)
cmake_options+=( \
  -DCMAKE_BUILD_TYPE=${cmake_build_type} \
  -DCMAKE_IGNORE_PATH="/usr/local/lib;/usr/local/include;/usr/local/bin" \
  -DCMAKE_INSTALL_PREFIX="${ROOTSYS}" \
  )

# Steering options.
add_steering_options

# Builtin options.
add_builtin_options

# Feature options.
add_feature_options

# Set compiler-specific options.
add_compiler_options

# Platform-specific options.
add_platform_options
####################################

##################
# Configure parallelism. *C*
ncores=$(ncores)

# For cmake --build:
[[ -n "${CMAKE_BUILD_PARALLEL_LEVEL}" ]] || \
  export CMAKE_BUILD_PARALLEL_LEVEL=${ncores}

# For ctest:
[[ -n "${CTEST_PARALLEL_LEVEL}" ]] || \
  export CTEST_PARALLEL_LEVEL=${ncores}
##################
####################################

####################################
# Configure, build, and install.
#
# Use ${blddir}/${pkgtarname} for in-source builds, ${blddir} otherwise.

${trace} cd "${blddir}" || \
  ssi_die "unable to cd to build directory ${blddir}"

##################
# Configure. *C*

${trace} cmake "${cmake_options[@]}" "${sourcedir}/root-${pkgdotver}" || \
  ssi_die "configure failed"
##################

##################
# Build. *C*

${trace} cmake --build "${blddir}" ${DEBUG_BUILD:+--verbose} || \
  ssi_die "build failed"
##################

##################
# Test. *C*
##################

##################
# Install. *C*

${trace} cmake --install "${blddir}" || ssi_die "install failed"
##################

ssi_info "finished building ${package} ${pkgver}"

####################################
# Pre-declaration processing. *C*

ssi_info "post-processing for relocatability"

fix_config_script \
  =ROOTSYS =ROOT_DIR \
  =FFTW_FQ_DIR \
  =LIBXML2_FQ_DIR \
  =GSL_FQ_DIR \
  =MYSQL_CLIENT_FQ_DIR \
  =NUMPY_FQ_DIR \
  =OPENBLAS_LIB \
  =POSTGRESQL_FQ_DIR \
  =PYLIB \
  =PYTHIA8_FQ_DIR \
  =PYTHON_INCLUDE \
  =PYTHON_LIB \
  =PYTHON_LIBDIR \
  =TBB_FQ_DIR \
  =XROOTD_LIB \
  =XROOTDROOT \
  -- \
  "${pkgdir}/bin/"{root-config,thisroot.sh} \
  "${pkgdir}/cmake/ROOTConfig-targets.cmake" \
  "${pkgdir}/config/Makefile.config" || exit
####################################

##################
# Declare and set up the installed package. *C*
#
# Add -c for current declaration if required.
${trace} declare_product -x -- || exit
##################

#################
# Verify installation and make distribution archive. *C*
#
# To customize verification, see is_built(), since it is also used by
# maybe_make_distribution(). See the documentation for those functions
# (above) for customization opportunities.

# Add a message to ssi_build_failed if appropriate.
${trace} is_built && ${trace} ssi_build_completed && \
  ssi_info "${package} is installed at ${pkgdir}" && \
  ${trace} maybe_make_distribution || \
  ssi_build_failed # *C*
########################################################################
# Nothing beyond this point.
########################################################################
